/*
Very simple GLUT program using bare OpenGL libraries to
demonstrate use of GL_ARB_shading_language_100.

Orion Sky Lawlor, olawlor@acm.org, 2006/10/05 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

#ifndef M_PI
# define M_PI           3.14159265358979323846  /* pi */
#endif


/* Variables updated by the GUI */
float spin_x, spin_y, zoom=1.0, specExp=100.0;
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0), light1Mat(1.0);

/* The thing to draw (sequence of glVertex calls) */
int model_mode=0; /* updated by listbox, below */
const char *model_mode_names[]={
	"Sphere", "Torus", "9 Points", "Triangle Stack",0
};

/* The way to draw it (glBegin mode) */
int render_mode=5; /* updated by listbox */
const char *render_mode_names[]={
	"GL_POINTS",
	"GL_LINES","GL_LINE_STRIP","GL_LINE_LOOP",
	"GL_TRIANGLES","GL_TRIANGLE_STRIP","GL_TRIANGLE_FAN",
	"GL_QUADS","GL_QUAD_STRIP",
	0
};
int render_mode_enums[]={
	GL_POINTS,
	GL_LINES,GL_LINE_STRIP,GL_LINE_LOOP,
	GL_TRIANGLES,GL_TRIANGLE_STRIP,GL_TRIANGLE_FAN,
	GL_QUADS,GL_QUAD_STRIP,
	0
};


void Exit(const char *where,const char *why) {
	fprintf (stderr, "FATALITY: OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

/************* Object we'll be drawing ***********/
class mainObject_t {
public:
	mainObject_t();
	void draw(void);
};


/** Set up a 3D model, stored in a vertex buffer */
mainObject_t::mainObject_t()
{
	/* nothing to set up here... */
}



/* These drawing routines are called from inside a
glBegin/glEnd pair.  They spit out a sequence of vertices... 
*/
/* Sphere drawn as a set of points */
void drawSphere(void)
{ 
	glColor4f(1,0,1,0.3); /* purple */
	double r=1.0;
	int nPiecesZ=20;
	double dlat=M_PI/nPiecesZ;
	for (double lat=-0.5*M_PI;lat+dlat<=0.5*M_PI;lat+=dlat)
	{
		double zL=r*sin(lat);
		double zH=r*sin(lat+dlat);
		double rL=r*sqrt(1-(zL*zL)); // xy-plane radius
		double rH=r*sqrt(1-(zH*zH));
		int nPieces=30; //1+(int)(rL*nPiecesZ);
		double dlon=2*M_PI/nPieces; /* radians-- size of a piece */
		/* Make a ring of points around the sphere */
		for (double lon=0;lon<=2*M_PI;lon+=dlon) {
			vec4 vL(rL*cos(lon),rL*sin(lon),zL,1.0);
			vec4 vH(rH*cos(lon),rH*sin(lon),zH,1.0);
			
			glNormal3fv(vL);
			glVertex4fv(vL);
			
			glNormal3fv(vH);
			glVertex4fv(vH);
		}
	}
	
}

/* Draw a torus */
void drawTorus(void) {
	glEnd(); /* because caller does glBegin... */
	glColor4f(1,1,0,0.3); /* yellow */
	glutSolidTorus(0.2,0.5,40,60);
	glBegin(GL_LINES); 
}

/* Draw a set of points */
void drawPoints(void) 
{
	const static vec4 verts[]={
		vec4(0.0,0,0,1),
		vec4(0.0,1,0,1),
		vec4(0.3,0,0,1),
		vec4(0.3,1,0,1),
		vec4(0.6,0,0,1),
		vec4(0.6,1,0,1),
		vec4(0.9,0,0,1),
		vec4(0.9,1,0,1),
		vec4(1.2,0,0,1),
	};
	glColor4f(1,0,0,0.5); /* Red points */
	for (int i=0;i<sizeof(verts)/sizeof(verts[0]);i++) 
		glVertex4fv(verts[i]);
}

/* Draw a stack of triangles */
void drawStack(void) 
{
	for (float bz=-0.25; bz<=+0.25;bz+=0.25) {
		glColor4f(1,0,0,1); /* Red triangles */
		glVertex4f(0,0,bz,1);
		glVertex4f(1,0,bz,1);
		
		glColor4f(1,1,0,1); /* Yellow vertex */
		glVertex4f(0,1,bz,1);

		glColor4f(0,1,0,1); /* Green triangles */
		float z=bz+0.1;
		glVertex4f(1,1,z,1);
		glVertex4f(1,0,z,1);
		glVertex4f(0,1,z,1);
	}
}

/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. 
*/
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}


/** Main drawing routine, called by main_display below */
void mainObject_t::draw(void)
{
	oglCheck("start draw");
	/* Set up programmable shader.
	  In a real program, you'd do all your glCreateProgramObjectARB 
	  calls at startup, instead of recompiling the program every frame... */
	oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	
	// Everything we draw from here will use our programmable shader...
	p->begin();
	
	
	/* Copy spin_x and spin_y into program uniforms */
	p->set("spin_x",spin_x);
	p->set("spin_y",spin_y);
	p->set("specExp",specExp);
	
	/**
	 Coordinate systems used here:
	
	Model coordinates, which are those passed in to glVertex
		modelMat transforms to world coordinates
	World coordinates, where lighting happens
		viewMat transforms to eye coordinates
	Eye coordinates, where the camera lies on the Z axis
		persMat transforms to clip coordinates
	Clip coordinates, onscreen position from -1 to +1 in X and Y
		the graphics hardware internally transforms clip coordinates to
	Window coordinates, which are just onscreen pixels.
	
	The vertex shader reads model coordinates in gl_Vertex
	and writes clip coordinates in gl_Position.
	*/
	mat4 modelMat=::modelMat; /* read rotation part of model matrix */
	modelMat[3].x+=0.3; /* shift model slightly in X direction*/
	p->set("modelMat",modelMat);
	p->set("lightDir",light1Mat*vec4(0,0,1,1));
	
	// Figure out where the camera is supposed to be
	double cameraDist=1.9; // Distance of camera from origin, in perspective matrix
	vec4 cameraLoc=vec4(0,0,cameraDist,1.0); // Location of camera after viewMat, and before perspective matrix
	mat4 viewInv=transpose(viewMat); // inverse is transpose for rotation matrix
	cameraLoc=viewInv*cameraLoc; // Location of camera in world coordinates
	p->set("cameraLoc",cameraLoc);
	
	// Perspective matrix
	mat4 persMat=mat4(1.0);
	persMat[0].x=zoom*win_h/(float)win_w; /* aspect ratio correction */
	persMat[1].y=zoom;
	persMat[2].z=-0.1; /* FIXME: set depth range in some smarter way... */
	persMat[2].w=-1.0; /* Flip Z axis, so -Z is deeper into screen (bigger W) */
	persMat[3].w=cameraDist;
	
	// Overall coordinate transform: perspective  view   model
	mat4 clipMat=persMat*viewMat*modelMat; 
	p->set("clipMat",clipMat);
	
	/* Finally, at long last, draw some friggin' objects! */
	glBegin(render_mode_enums[render_mode]);
	switch (model_mode) {
	case 0: drawSphere(); break;
	case 1: drawTorus(); break;
	case 2: drawPoints(); break;
	case 3: drawStack(); break;
	};
	glEnd();
	
	/* Stop using programmable shader */
	p->end();
	delete p;
	
	/* This line causes the world to repeatedly redraw itself, using 100% of the CPU. */
	glutPostRedisplay();
	oglCheck("end draw");
}


/******** Program Initialization *********/
mainObject_t *mainObject=0; /**< Big object we're trying to draw */


/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	Check("Entering display routine");
	// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Set up a little error-prone OpenGL state:
	glDisable(GL_POLYGON_SMOOTH); /* smooth edges of polygons */
	glDisable(GL_BLEND); /* alpha blending */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glPointSize(10.0); /* diameter in pixels */
	glLineWidth(10.0); /* thickness in pixels */
	
	// Draw stuff
	Check("About to draw");
	mainObject->draw();
	Check("Done drawing");
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	Check("Before glViewport");
	win_w=w; win_h=h-50; //<- hack: leave room at top for GLUI spinner boxes
	if (win_w<0) win_w=0; // <- Prevents "invalid value" error for small windows...
	if (win_h<0) win_h=0;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
	Check("After glViewport");
}

void make_list(GLUI *glui,const char *title,int *mode_ptr,const char **names)
{
	GLUI_Listbox *list = new GLUI_Listbox( glui, title, mode_ptr);
	for (int i=0;names[i];i++) 
		list->add_item(i,names[i]);
	list->do_selection(*mode_ptr);
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");
	
	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();

	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	// Make an object to draw
	mainObject=new mainObject_t;
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_TOP);

	GLUI_Spinner *spinner_x = new GLUI_Spinner( glui, "X:", &spin_x);
	spinner_x->set_float_limits(0.0, 1.0);
/*	GLUI_Spinner *spinner_y = new GLUI_Spinner( glui, "Y:", &spin_y);
	spinner_y->set_float_limits(0.0, 1.0);
*/
	GLUI_Spinner *spin_zoom = new GLUI_Spinner( glui, "Zoom:", &zoom);
	spin_zoom->set_float_limits(0.0, 10.0);
	GLUI_Spinner *spin_spec = new GLUI_Spinner( glui, "SpecExp:", &specExp);
	spin_spec->set_float_limits(0.0, 1000.0);

	new GLUI_Column(glui);
	
	make_list(glui,"Model:",&model_mode,model_mode_names);
	make_list(glui,"glBegin:",&render_mode,render_mode_names);
	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"Light:",&light1Mat[0][0]);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
